/*
Johann Alvarez
7/26/2013

Research Project
===========================
Simple Five Card Stud Poker game with slight variations. This program is meant to show how the AI
'learns' from an opponent's strategy and uses a different counter-strategy depending on the
opponent's bluff percentage, so it won't be like real poker. The betting is only significant for 
the round itself, not for the entire game. The agents are both given $150 at the beginning of each round.
There will be 11 rounds. The cards at the end of each round are shown to both the player and the AI. 
Both the player and the AI have three choices to make in each round: 1)All in, 2)A Third Bet, 3)Fold. 
A round can go up to three turns, on account of the 'thirds' bet. A bluff counts as a player going all 
in with a hand that is weaker than a straight, or if the player calls the AI's third bet up until the 
final turn and then folds. For every round, whoever wins the round gets a point, but only if their opponent
didn't fold in the first turn. Whoever ends the most points at the end of the game wins.

*/
#include <iostream>
#include <cstdlib>
#include <ctime>
#include <string>
using namespace std;

string suits[4] = {"Clubs", "Spades", "Diamonds", "Hearts"};
string ranks[13] = {"2", "3", "4", "5", "6", "7", "8", "9", "10", "Jack", "Queen", "King", "Ace"};

double bluffPercentage;

class Card {
	public:
		int suit;
		int rank;
};

class Deck {
	public:
		Deck(){
            for (int i = 0; i < 4; i++){
                for (int j = 0; j < 13; j++){
                    cards[i*13+j].suit = i;
                    cards[i*13+j].rank = j;
                }
            }
        }

        void print(){
            cout << "Printing the deck..." << endl;
            for (int i = 0; i < 52; i++)
                cout << ranks[cards[i].rank] << " of " << suits[cards[i].suit] << endl;
            cout << endl;
        }

        void shuffle(){
            top = 51;
            int x;
            Card tempCard;
            for (int i = 0; i < 4; i++){
                for (int j = 0; j < 13; j++) {
                    cards[i*13+j].suit = i;
                    cards[i*13+j].rank = j;
                }
            }
            cout << "Shuffling the cards and dealing..." << endl << endl;
            for (int i = 0; i < 52; i++){
                x = rand() % 52;
                tempCard = cards[i];
                cards[i] = cards[x];
                cards[x] = tempCard;
            }
        }

        Card dealOneCard() {
            top--;
            return cards[top+1];
        }

    private:
		int top;
		Card cards[52];
};

int compareCards(const void *card1, const void *card2) 
{
    return (*(Card *)card1).rank - (*(Card *)card2).rank;
}

class Player {
	public:
		int money;
		Card cards[5];
};

class AI {
	public:
		int money;
		Card cards[5];
		int choice;
		void strategy(int c, double bluf)
		{
			// if player has bluff percentage greater than 60 and they go all in, AI goes all in as well.
			if (c == 1 && bluf >= 60)
			{
				choice = 1;
			}
			else if (c == 1 && bluf < 60)
			{
				choice = 3;
			}
			else
			{
				choice = 2;
			}
		}
};

int getScore(Card hand[]){
	qsort(hand,5,sizeof(Card),compareCards);
	int straight, flush, three, four, full, pairs, high;
	int k;

	straight = flush = three = four = full = pairs = high = 0;
	k = 0;

	/*checks for flush*/
	while (k<4&&hand[k].suit==hand[k+1].suit)
		k++;
	if (k==4)
		flush = 1;

            /* checks for straight*/
            k=0;
            while (k<4&&hand[k].rank==hand[k+1].rank-1)
                k++;
            if (k==4)
                straight = 1;

            /* checks for fours */
            for (int i=0;i<2;i++){
                k = i;
                while (k<i+3&&hand[k].rank==hand[k+1].rank)
                    k++;
                if (k==i+3){
                    four = 1;
                    high = hand[i].rank;
                }
            }

            /*checks for threes and fullhouse*/
            if (!four){
                for (int i=0;i<3;i++){
                    k = i;
                    while (k<i+2&&hand[k].rank==hand[k+1].rank)
                        k++;
                    if (k==i+2){
                        three = 1;
                        high=hand[i].rank;
                        if (i==0){
                            if (hand[3].rank==hand[4].rank)
                                full=1;
                        }
                        else if(i==1){
                            if (hand[0].rank==hand[4].rank)
                                full=1;
                        }
                        else{
                            if (hand[0].rank==hand[1].rank)
                                full=1;
                        }
                    }
                }
            }

            if (straight&&flush)
                return 170 + hand[4].rank;
            else if(four)
                return 150 + high;
            else if(full)
                return 130 + high;
            else if(flush)
                return 110;
            else if(straight)
                return 90 + hand[4].rank;
            else if(three)
                return 70 + high;

            /* checks for pairs*/
            for (k=0;k<4;k++) {
                if (hand[k].rank==hand[k+1].rank){
                    pairs++;
                    if (hand[k].rank>high)
                        high = hand[k].rank;
                }
			}

            if (pairs==2)
                return 50 + high;
            else if(pairs)
                return 30 + high;
            else
                return hand[4].rank;
        }

int main()
{
	srand(time(NULL));
	int pot = 0;
	int playerPoints = 0, aiPoints = 0, finalPoints = 0;
	int numRounds;
	int choice;

	bluffPercentage = 0.0;
	
	// print welcome message
	cout << "Welcome to the Crazy Five Card Poker Game! (Research Project)" << '\n';
	cout << "==============================================================" << "\n\n";

	// create player and AI
	Player player;
	AI ai;

	// create a new deck
	Deck deck;

	// print cards (will be commented out later)
	//deck.print();

	cout << "Ok, let's start the game!" << "\n\n";

	for (int i = 0; i < 11; i++)
	{
		player.money = 150;
		ai.money = 150;
			
		// shuffle the deck
		deck.shuffle();

		//while (player.money != 0 && ai.money != 0)
		//{
			bool fold = false;
			bool thirdbet = false;
			bool allin = false;

			int bluffing = 0;
	
			cout << "Player Money = $" << player.money << '\n';
			cout << "AI Money = $" << ai.money << "\n\n";

			// deal the cards to player and AI, alternating
			for (int i = 0; i < 5; i++) {
				player.cards[i] = deck.dealOneCard();
				ai.cards[i] = deck.dealOneCard();
			}
	
			cout << "Here are the Player's cards:" << endl;
			for (int i = 0; i < 4; i++)
				cout << ranks[player.cards[i].rank] << " of " << suits[player.cards[i].suit] << ", ";
			cout << ranks[player.cards[4].rank] << " of " << suits[player.cards[4].suit] << "\n\n";

			cout << "Here are the AI's cards:" << endl;
			for (int i = 0; i < 4; i++)
				cout << ranks[ai.cards[i].rank] << " of " << suits[ai.cards[i].suit] << ", ";
			cout << ranks[ai.cards[4].rank] << " of " << suits[ai.cards[4].suit] << "\n\n";
	
			do {
				// ask the player what they would like to do
				cout << "Player, what would you like to do? " << '\n';
				cout << "1) All in " << '\n';
				cout << "2) 1/3 bet " << '\n';
				cout << "3) Fold " << '\n';
				cout << "Make your selection ->";

				cin >> choice;
				cout << '\n';

				switch (choice)
				{
					case 1: cout << "Player goes all in!\n" <<'\n';
							allin = true;
							pot += player.money;
							player.money -= 150;
							break;
					case 2: cout << "Player makes 1/3 bet.\n" << '\n';
							thirdbet = true;
							pot += 50;
							player.money -= 50;
							break;
					case 3: cout << "Player folds!\n" << '\n';
							fold = true;
							break;
				}

				ai.strategy(choice, bluffPercentage);
				if (ai.choice == 1)
				{
					pot += ai.money;
					ai.money = 0;
				}
				else if (ai.choice == 2)
				{
					pot += 50;
					ai.money -= 50;
				}
			
				cout << "Player Money = $" << player.money << '\n';
				cout << "AI Money = $" << ai.money << "\n\n";

			} while (player.money != 0 && ai.money != 0 && fold != true);

			// score each hand
			playerPoints = getScore(player.cards);
			aiPoints = getScore(ai.cards);

			cout << "Player has " << playerPoints << " points." << '\n';
			cout << "AI has " << aiPoints << " points." << '\n';

			if (playerPoints > aiPoints && fold != true)
			{
				finalPoints = playerPoints;
				cout << "Player";
			}
			else
			{
				finalPoints = aiPoints;
				cout << "AI";
			}

			cout << " wins $" << pot << " with ";
			if (finalPoints<30)
				cout << "HIGH CARD";
			else if (finalPoints<50)
				cout << "SINGLE PAIR";
			else if (finalPoints<70)
				cout << "TWO PAIRS";
			else if (finalPoints<90)
				cout << "THREE OF A KIND";
			else if(finalPoints<110)
				cout << "STRAIGHT";
			else if(finalPoints<130)
				cout << "FLUSH";
			else if(finalPoints<150)
				cout << "FULL HOUSE";
			else if(finalPoints<170)
				cout << "FOUR OF A KIND";
			else
				cout << "STRAIGHT FLUSH";
	
			cout << "\n\n";

			if (player.money = 0 && finalPoints > 100) 
			{
				bluffing++;
			}

			bluffPercentage = bluffing / i;
		
	}

	cout << "Thank you for playing! Have a great day!" << endl;

	char pause;
	cin >> pause;
	return 0;
}